# TODO(nopdive): This module requires a memory review.

import logging
import time
import weakref
from abc import ABC
from typing import TYPE_CHECKING, Any, Callable, Iterator, Optional, Sequence

import numpy as np

from ..._parser import TokenParser
from ..._schema import (
    EngineCallResponse,
    EngineOutput,
    GenToken,
    GuidanceEngineMetrics,
    LLGrammar,
)
from ...registry import get_exchange
from ..._utils import log_cleanup, log_init, softmax, to_utf8_or_bytes_string
from ...visual import (
    ExecutionCompletedMessage,
    ExecutionStartedMessage,
    GuidanceMessage,
    OutputRequestMessage,
    MetricMessage,
)
from ._state import EngineState
from ._tokenizer import Tokenizer

if TYPE_CHECKING:
    from .._base._model import Model

logger = logging.getLogger(__name__)


def _engine_cleanup(msg_recv: Callable[[GuidanceMessage], None], log_msg: str):
    get_exchange().unsubscribe(msg_recv)
    log_cleanup(log_msg)


def _wrapped_msg_recv(engine_weak_ref: weakref.ref) -> Callable[[GuidanceMessage], None]:
    def closure(message):
        return _msg_recv(engine_weak_ref, message)

    return closure


def _msg_recv(engine_weakref: weakref.ReferenceType, message: GuidanceMessage) -> None:
    # NOTE(nopdive): This is run on a background thread.

    engine = engine_weakref()
    if engine is None:
        return

    if isinstance(message, MetricMessage):
        return

    # logger.debug(f"ENGINE({id(engine)}):msg_recv:{message}")
    if isinstance(message, ExecutionStartedMessage):
        pass
    elif isinstance(message, ExecutionCompletedMessage) and message.is_err:
        pass
    elif isinstance(message, ExecutionCompletedMessage):
        pass
    elif isinstance(message, OutputRequestMessage):
        pass


class Engine(ABC):
    """The engine owns the inference computation and is used/created by the Model class.

    Engine objects represent the expensive parts of inference. While Model objects are cheap and do not
    need to know about the tokenizer or the model parameters, Engine objects know about both. Many
    Model objects can reference a single Engine object. Engine objects can also be hidden behind a
    Server so a single server can serve many clients' model objects through a single Engine object.
    """

    def __init__(self, tokenizer: Tokenizer, compute_log_probs=False, enable_backtrack=True, enable_ff_tokens=True,
                 enable_monitoring=True, **kwargs):
        from ...registry import get_monitor

        self.tokenizer = tokenizer
        self.compute_log_probs = compute_log_probs
        self._enable_backtrack = enable_backtrack
        self._enable_ff_tokens = enable_ff_tokens
        self._enable_monitoring = enable_monitoring
        self._top_k = kwargs.get("top_k", 5)

        # TODO(nopdive): Remove on refactor.
        self.metrics = GuidanceEngineMetrics()

        if self._enable_monitoring:
            # Idempotent start
            _ = get_monitor()

        msg_recv = _wrapped_msg_recv(weakref.ref(self))
        get_exchange().subscribe(msg_recv)

        weakref.finalize(
            self,
            _engine_cleanup,
            msg_recv,
            f"engine({id(self)})",
        )
        log_init(f"engine({id(self)})")

    # These need to be properties because once an Engine is started, you can't change their behavior.
    @property
    def enable_backtrack(self):
        return self._enable_backtrack

    @property
    def enable_ff_tokens(self):
        return self._enable_ff_tokens

    @property
    def enable_monitoring(self):
        return self._enable_monitoring

    def get_chat_template(
        self,
    ):  # TODO [HN]: Add more logic here...should we instantiate class here? do we even need to?
        return (
            self.tokenizer.chat_template()
        )  # Instantiate the class before returning to client for now

    def reset_metrics(self):
        self.metrics = GuidanceEngineMetrics()

    def __call__(
        self,
        state: EngineState,
        grammar: LLGrammar,
        ensure_bos_token: bool = True,
        echo: bool = True,
    ) -> Iterator[EngineCallResponse]:
        """Main entry point for the inference-parser loop. Yields EngineCallResponse objects as
        the parser advances through the grammar.

        Parameters
        ----------
        prompt : str or Parser
            This is represents the current state of a guidance parser that will be extended
            using the passed grammar. If a string is given then we assume the previous parser
            state is just a fixed string prompt, if a full Parser is given then we extend that
            parser by appending the new grammar to the parser's current grammar and then
            inferencing the model. (TODO: implement full parser extension support)
        grammar: Function
            Grammar (RawFunction or GrammarFunction) used to extend the prompt.
        ensure_bos_token: bool
            Ensures that the prompt ends with the BOS token.
        """
        # TODO: Pass these to get_logits
        # images = state.images
        # audio = state.audio
        # videos = state.videos

        parser = TokenParser(
            grammar,
            tokenizer=self.tokenizer,
            prompt=state.prompt.encode("utf-8"),
            ensure_bos_token=ensure_bos_token,
            enable_backtrack=self.enable_backtrack,
            enable_ff_tokens=self.enable_ff_tokens,
        )

        has_get_logits = True
        engine_output = None
        logits_lat_ms = 0
        delayed_bytes = b""
        delayed_engine_outputs: list[EngineOutput] = []
        while not parser.done():
            t0 = time.time()

            tokens, mask_fut, backtrack = parser.advance(engine_output)

            # Note that has_pending_stop implies that the response is a stop response,
            # but the converse is not true. We can therefore avoid some (but not all)
            # unnecessary calls to get_logits on the final iteration.
            has_pending_stop = parser.has_pending_stop()

            if has_get_logits and not has_pending_stop:
                try:
                    t0 = time.time()
                    logits = self.get_logits(token_ids=tokens)
                    logits_lat_ms = (time.time() - t0) * 1000
                except NotImplementedError:
                    # Permanently fall-back to get_next_token if get_logits is not implemented
                    has_get_logits = False
                    logits = None
                    logits_lat_ms = 0
            else:
                logits = None

            # Important: don't wait on this future until after getting the logits;
            # this allows the mask to be built concurrently with model inference
            mask, ll_response = mask_fut.result()

            engine_response = ll_response.progress.to_engine_call_response()
            engine_response.backtrack = backtrack
            if engine_output:
                engine_response.engine_outputs.append(engine_output)

            # NOTE (loc): Temporary solution to quickly check which segments are generated and which are force-forwarded to animate visualizations on the UI
            # These tokens in chunk will not be used for final visualization
            # TODO: This should be handled by the interpreter
            if echo and engine_response.new_bytes:
                try:
                    _new_bytes = delayed_bytes + engine_response.new_bytes
                    _tokens = parser.tokenizer.encode(_new_bytes)
                    delayed_bytes = b""
                except UnicodeDecodeError:
                    # similar to what we did in _run_stateless function, if we could not decode current bytes
                    # we will delay until we can decode them
                    delayed_bytes += engine_response.new_bytes
                    if engine_output:
                        engine_response.engine_outputs.pop()
                        delayed_engine_outputs.append(engine_output)

                if not delayed_bytes:
                    ff_token_start_idx = 1
                    if engine_output is None and len(delayed_engine_outputs) == 0:
                        ff_token_start_idx = 0
                    elif (
                        engine_output.issued_token.token_id == _tokens[0]
                        and len(delayed_engine_outputs) == 0
                    ):
                        # this is generated
                        engine_response.generated_bytes = parser.tokenizer.decode([_tokens[0]])
                        engine_output.issued_token.is_generated = True
                        engine_response.generated_tokens.append(engine_output.issued_token)
                    else:
                        # handle delayed bytes
                        engine_outputs = (
                            delayed_engine_outputs + [engine_output] if engine_output else []
                        )
                        engine_output_tokens = [e.issued_token.token_id for e in engine_outputs]

                        generated = to_utf8_or_bytes_string(
                            parser.tokenizer.decode(engine_output_tokens)
                        )
                        force_forwarded = _new_bytes.decode("utf-8")

                        if force_forwarded.startswith(generated):
                            engine_output_tokens = np.array(engine_output_tokens)
                            ff_tokens = np.array(_tokens)

                            # check if engine_output_tokens in ff_tokens
                            _idx = -1
                            for _i in range(0, len(ff_tokens) - len(engine_output_tokens) + 1):
                                if np.array_equal(
                                    engine_output_tokens,
                                    ff_tokens[_i : _i + len(engine_output_tokens)],
                                ):
                                    _idx = _i + len(engine_output_tokens)
                                    break

                            if _idx < 0:
                                ff_token_start_idx = 0
                            else:
                                # all previous tokens before _idx are generated
                                engine_response.generated_bytes = parser.tokenizer.decode(
                                    ff_tokens[:_idx]
                                )
                                idx_in_engine_output_tokens = 0
                                for _i in range(_idx):
                                    matching_engine_output = None
                                    if (
                                        _tokens[_i]
                                        == engine_output_tokens[idx_in_engine_output_tokens]
                                    ):
                                        matching_engine_output = engine_outputs[
                                            idx_in_engine_output_tokens
                                        ]
                                        idx_in_engine_output_tokens += 1
                                    engine_response.generated_tokens.append(
                                        GenToken(
                                            token_id=_tokens[_i],
                                            prob=(
                                                1.0
                                                if not matching_engine_output
                                                else matching_engine_output.issued_token.prob
                                            ),
                                            text=(
                                                parser.tokenizer.decode([_tokens[_i]])
                                                if not matching_engine_output
                                                else matching_engine_output.issued_token.text
                                            ),
                                            latency_ms=(
                                                0.0
                                                if not matching_engine_output
                                                else matching_engine_output.issued_token.latency_ms
                                            ),
                                            is_generated=True,
                                        )
                                    )
                                ff_token_start_idx = _idx
                        else:
                            ff_token_start_idx = 0

                    if len(_tokens[ff_token_start_idx:]):
                        engine_response.force_forwarded_bytes = parser.tokenizer.decode(
                            _tokens[ff_token_start_idx:]
                        )
                        for _token in _tokens[ff_token_start_idx:]:
                            engine_response.force_forwarded_tokens.append(
                                GenToken(
                                    token_id=_token,
                                    prob=1.0,
                                    text=to_utf8_or_bytes_string(
                                        parser.tokenizer.decode([_token])
                                    ),
                                    latency_ms=0,
                                    is_force_forwarded=True,
                                )
                            )

                    delayed_engine_outputs = []
            elif not echo and engine_response.new_bytes:
                # do not collect tokens-metrics if echo is disabled
                engine_response.generated_bytes = engine_response.new_bytes
                engine_response.generated_tokens.clear()

            # process engine_response
            # NOTE (loc): We should not yield the engine_response if new_bytes are invalid utf-8 bytes
            # delayed bytes should be handled here in the engine
            yield engine_response

            if ll_response.stop:
                assert mask is None
                # May raise an exception if the parser is in an bad state!
                parser.cleanup()
                # Ensure we break AFTER yielding the final response
                break

            # If there was a pending stop, we should have broken out of the loop
            assert not has_pending_stop

            # Help the type checker: assert that everything we need to get the next token is not None
            assert mask is not None
            assert ll_response.temperature is not None

            can_finish_early = parser.is_accepting() and self.tokenizer.eos_token_id is not None

            if can_finish_early:
                # Type checker needs some help
                assert self.tokenizer.eos_token_id is not None
                # Should be equivalent to parser.is_accepting()
                assert mask[self.tokenizer.eos_token_id]
                # Whenever we are in an accepting state, we will allow the model to generate whatever it wants
                # but we will treat any "illegal" tokens as EOS, allowing the model to finish gracefully.
                # Hence, mask must be None
                mask_for_sampling = None
            else:
                mask_for_sampling = mask

            engine_output = self.get_next_token_with_top_k(
                logits=logits,
                logits_lat_ms=logits_lat_ms,
                token_ids=tokens,
                mask=mask_for_sampling,
                temperature=ll_response.temperature,
                k=self._top_k,
                force_return_unmasked_probs=echo,
            )

            if can_finish_early and not mask[engine_output.issued_token.token_id]:
                # Type checker needs some help
                assert self.tokenizer.eos_token_id is not None
                engine_output.issued_token.token_id = self.tokenizer.eos_token_id

    def get_next_token_with_top_k(
        self,
        logits: Optional[np.ndarray],
        logits_lat_ms: Optional[float],
        token_ids: list[int],
        mask: Optional[bytes],
        temperature: float,
        k: int = 5,
        force_return_unmasked_probs: bool = False,
    ) -> EngineOutput:
        """Get the next token and associated top-k tokens from the engine.

        Parameters
        -------
        logits : Optional[np.ndarray]
            The logits for the current token ids in the sequence.
            If None, the model will call get_logits to get the logits.
        logits_lat_ms: Optional[float]
            The time taken to compute the logits.
            If logits is None, the model will call get_logits to measure the time.
        token_ids : list[int]
            The current token ids in the sequence.
        mask : Optional[bytes]
            The mask to apply to the logits.
        temperature : float
            The temperature to apply to the logits.
        k : int
            The number of top-k tokens to return.
        force_return_unmasked_probs: bool
            If True, the top-k unmasked probabilities will be returned.

        Returns
        -------
        EngineOutput
            The output from the model.
        """

        if logits is None:
            t0 = time.time()
            try:
                logits = self.get_logits(token_ids)
            except NotImplementedError:
                # fallback to orignal get_next_token method
                _t0 = time.time()
                token_id = self.get_next_token(
                    token_ids=token_ids,
                    mask=mask,
                    temperature=temperature,
                )
                _lat = (time.time() - _t0) * 1000

                _issued_token = GenToken(
                    token_id=token_id,
                    prob=1.0,
                    text=to_utf8_or_bytes_string(self.tokenizer.decode([token_id])),
                    latency_ms=_lat,
                    is_generated=True,
                )

                return EngineOutput(
                    issued_token=_issued_token,
                    top_k=[_issued_token],
                    masked_top_k=[_issued_token] if mask is not None else None,
                )
            lat_ms = (time.time() - t0) * 1000
        else:
            lat_ms = logits_lat_ms

        def get_top_k(_probs: np.ndarray, _k: int = 5) -> list[GenToken]:
            top_k_indices = np.argsort(_probs)[::-1][:_k]
            top_k_probs = _probs[top_k_indices]

            return [
                GenToken(
                    token_id=token,
                    prob=prob,
                    text=to_utf8_or_bytes_string(self.tokenizer.decode([token])),
                    latency_ms=lat_ms,
                    is_generated=True,
                )
                for token, prob in zip(top_k_indices, top_k_probs)
                if prob > 0
            ]

        # compute top-k without masking
        probs = (
            softmax(np.array(logits))
            if temperature < 0.0001
            else softmax(np.array(logits) / temperature)
        )

        top_k: list[GenToken] = []
        if force_return_unmasked_probs:
            top_k = get_top_k(probs, k)

        # compute top-k with masking
        masked_top_k: list[GenToken] = []
        if mask is not None:
            # shift logits to [0 - max] range first and apply mask
            masked_logits = (logits - np.min(logits)) * np.frombuffer(mask, dtype=np.uint8)
            masked_probs = (
                softmax(masked_logits)
                if temperature < 0.0001
                else softmax(masked_logits / temperature)
            )
            masked_top_k = get_top_k(masked_probs, k)

        if temperature < 0.0001:
            if len(masked_top_k) > 0:
                issued_token = masked_top_k[0]
            else:
                if len(top_k) == 0:
                    top_k = get_top_k(probs, k)
                issued_token = top_k[0]
        else:
            # we need to sample from the probabilities
            if mask is None:
                sampled_index = np.random.choice(len(probs), p=probs)
                sampled_prob = probs[sampled_index]
            else:
                sampled_index = np.random.choice(len(masked_probs), p=masked_probs)
                sampled_prob = masked_probs[sampled_index]

            issued_token = GenToken(
                token_id=sampled_index,
                prob=sampled_prob,
                text=to_utf8_or_bytes_string(self.tokenizer.decode([sampled_index])),
                latency_ms=lat_ms,
                is_generated=True,
            )

        output = EngineOutput(
            issued_token=issued_token,
            top_k=top_k,
            masked_top_k=None if not masked_top_k else masked_top_k,
            is_backtracked=False,
        )

        return output

    def get_next_token(
        self, token_ids: list[int], mask: Optional[bytes], temperature: float
    ) -> int:
        # Prefer to implement get_logits over get_next_token as it allows for concurrent mask computation
        raise NotImplementedError

    def get_logits(self, token_ids: list[int]) -> np.ndarray:
        # Prefer to implement get_logits over get_next_token as it allows for concurrent mask computation
        raise NotImplementedError

    def get_per_token_topk_probs(self, token_ids: list[int], top_k: int = 5) -> list[GenToken]:
        """Get the top-k probabilities for each token in the sequence."""
        raise NotImplementedError

    def sample_with_temperature(
        self, logits: np.ndarray, mask: Optional[bytes], temperature: float
    ) -> int:
        if mask is not None:
            logits += np.frombuffer(mask, dtype=np.uint8)
        if temperature < 0.0001:
            return int(np.argmax(logits))
        # Get probabilities from softmax
        probabilities = softmax(logits / temperature)
        # Sample an index based on the probabilities
        sampled_index = np.random.choice(len(logits), p=probabilities)
        return sampled_index

    def _report_failed_match(self, prompt):
        """Note that this can be overridden by subclasses that have more likely reasons than a bug in the token set (like remote models)."""
        return Exception(
            "We can't consume any more tokens, but we are not yet done! Perhaps your model's token set is incomplete? This happened after the prompt:"
            + str(prompt[-40:])
        )


class ConstraintException(Exception):
    def __init__(self, *args, **kwargs):
        self.prompt = kwargs.pop("prompt", None)
        self.data = kwargs.pop("data", None)
        super().__init__(*args, **kwargs)





import json
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
)

from ._parser import ByteParser, ByteParserException
from ._schema import JsonGrammar, LarkGrammar, LLGrammar
from .trace import OutputAttr

if TYPE_CHECKING:
    from .models._base import Client, State

# to support the embedding of guidance functions inside Python f-strings we use tags with these delimiters
tag_start = "{{G|"  # start of a call tag
tag_end = "|G}}"  # end of a call tag
_tag_pool: dict[str, Union["Function", "GrammarNode"]] = (
    {}
)  # the Functions or GrammarNodes associated with the tags
_tag_pattern = re.compile(
    re.escape(tag_start) + r"([^\|]+)" + re.escape(tag_end)
)  # the pattern for matching call tags


def _parse_tags(s: str) -> Union["GrammarNode", "Function"]:
    parts = cast(list[str], _tag_pattern.split(s))
    obj: GrammarNode = LiteralNode(parts.pop(0))
    is_tag = True
    for part in parts:
        if is_tag:
            obj += _tag_pool[part]
        else:
            obj += LiteralNode(part)
        is_tag = not is_tag
    return obj


class Tagged:
    def __str__(self):
        """Creates a string tag that can be used to retrieve this object."""

        # save the call in our call pool, ready to be run when it is attached to an LM object
        str_id = str(id(self))
        if str_id not in _tag_pool:
            _tag_pool[str_id] = self

        # return a string representation of this call so it can be combined with other strings/calls
        return tag_start + str_id + tag_end


class Match:
    def __init__(self, captures, log_probs, partial):
        self.captures = captures
        self.log_probs = log_probs
        self.partial = partial

    def __getitem__(self, key):
        return self.captures[key]

    def __len__(self):
        return len(self.captures)

    def __bool__(self):
        return True

    def __str__(self):
        return str(self.captures)

    def __repr__(self):
        return (
            "<guidance.Match object; captures="
            + str(self.captures)
            + "; partial="
            + str(self.partial)
            + ">"
        )


class StatefulException(Exception):
    """This is raised when we try and use the state of a grammar object like it was a live model.

    Note that eventually it would be nice to support stateful parser/grammar constructs directly, but
    such "parser combinators" cannot be run effciently in Python. So we use a traditional parser and
    grammar separation (hence the need for this exception)."""

    pass


@dataclass
class Function(Tagged):
    name: str = field(init=False)
    f: Callable
    args: tuple[Any, ...]
    kwargs: dict[str, Any]

    def __post_init__(self):
        self.name = self.f.__name__

    def __call__(self, model):
        model = self.f(model, *self.args, **self.kwargs)
        if model is None:
            raise Exception(
                f"The guidance function `{self.f.__name__}` did not return a model object! You need to return an updated model object at the end of your guidance function."
            )
        return model

    def __add__(self, other):
        if not isinstance(other, (str, GrammarNode, Function)):
            return NotImplemented

        if isinstance(other, str):
            other = _parse_tags(other)

        if isinstance(other, GrammarNode) and other.is_null:
            return self

        def __add__(model):
            return self(model) + other

        return Function(__add__, [], {})

    def __radd__(self, other):
        if not isinstance(other, (str, GrammarNode, Function)):
            return NotImplemented

        if isinstance(other, str):
            other = _parse_tags(other)

        if isinstance(other, GrammarNode) and other.is_null:
            return self

        def __radd__(model):
            return self(model + other)

        return Function(__radd__, [], {})


S = TypeVar("S", bound="State")


class ASTNode(ABC):
    @abstractmethod
    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        pass

    def simplify(self) -> "ASTNode":
        return self


@dataclass
class RoleStart(ASTNode):
    role: str

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client._role_start(state, self, **kwargs)


@dataclass
class RoleEnd(ASTNode):
    role: str

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client._role_end(state, self, **kwargs)


@dataclass
class ImageBlob(ASTNode):
    data: str

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.image_blob(state, self, **kwargs)


@dataclass
class ImageUrl(ASTNode):
    url: str

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.image_url(state, self, **kwargs)


@dataclass
class AudioBlob(ASTNode):
    data: str

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.audio_blob(state, self, **kwargs)


class GenAudio(ASTNode):
    def __init__(self, kwargs: dict[str, Any]):
        self.kwargs = kwargs

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.gen_audio(state, self, **kwargs)


@dataclass(frozen=True)
class GrammarNode(Tagged, ASTNode):

    @property
    def is_null(self) -> bool:
        """
        If this returns true, then this node matches empty string and empty string only.
        """
        return False

    @property
    def is_terminal(self) -> bool:
        """
        If this returns true, then this node will be compiled down to a regular expression.
        It cannot be recursive.
        """
        return all(child.is_terminal for child in self.children())

    def simplify(self) -> "GrammarNode":
        return self

    def children(self) -> Sequence["GrammarNode"]:
        return ()

    def __repr__(self) -> str:
        return self.lark_str()

    def __add__(self, other) -> "GrammarNode":
        if not isinstance(other, (str, GrammarNode)):
            return NotImplemented

        if isinstance(other, str):
            other = _parse_tags(other)

        if self.is_null:
            return other

        if isinstance(other, Function):
            return other.__radd__(self)

        if other.is_null:
            return self

        return JoinNode((self, other))

    def __radd__(self, other) -> "GrammarNode":
        if not isinstance(other, (str, GrammarNode)):
            return NotImplemented

        if isinstance(other, str):
            other = _parse_tags(other)

        if self.is_null:
            return other

        if isinstance(other, Function):
            return other.__add__(self)

        if other.is_null:
            return self

        return JoinNode((other, self))

    def __getitem__(self, key):
        raise StatefulException("GrammarNodes can't access state!")

    def match(
        self,
        byte_string: Union[str, bytes],
        allow_partial: bool = False,
        raise_exceptions: bool = False,
    ) -> Union[Match, None]:
        if isinstance(byte_string, str):
            byte_string = byte_string.encode()
        parser = ByteParser(self.ll_grammar())

        try:
            parser.consume_bytes(byte_string)
            if not allow_partial:
                parser.force_done()
        except ByteParserException:
            if raise_exceptions:
                raise
            else:
                return None

        if not allow_partial and not parser.matched():
            return None

        if parser.matched():
            parser.force_done()

        return Match(*parser.get_captures(), partial=not parser.matched())  # type: ignore[misc]

    def forced_prefix(self) -> str:
        parser = ByteParser(self.ll_grammar())
        return parser.bytes.decode("utf-8", errors="ignore")

    def ll_grammar(self) -> LLGrammar:
        return LLSerializer().serialize(self)


@dataclass(frozen=True)
class LiteralNode(GrammarNode):
    value: str

    @property
    def is_null(self) -> bool:
        return self.value == ""

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.text(state, self, **kwargs)


@dataclass(frozen=True)
class RegexNode(GrammarNode):
    regex: Optional[str]

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.regex(state, self, **kwargs)


@dataclass(frozen=True)
class SelectNode(GrammarNode):
    alternatives: tuple[GrammarNode, ...]

    @property
    def is_null(self) -> bool:
        return all(alt.is_null for alt in self.alternatives)

    def simplify(self) -> "GrammarNode":
        if self.is_null:
            return LiteralNode("")
        alternatives = tuple(alt.simplify() for alt in self.alternatives if not alt.is_null)
        if len(alternatives) == 1:
            node = alternatives[0]
        else:
            node = SelectNode(alternatives)

        if any(alt.is_null for alt in self.alternatives):
            return RepeatNode(node, 0, 1)
        return node

    def children(self) -> Sequence["GrammarNode"]:
        return self.alternatives

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.select(state, self, **kwargs)


@dataclass(frozen=True)
class JoinNode(GrammarNode):
    nodes: tuple[GrammarNode, ...]

    @property
    def is_null(self) -> bool:
        return all(node.is_null for node in self.nodes)

    def simplify(self) -> "GrammarNode":
        if self.is_null:
            return LiteralNode("")
        nodes = tuple(node.simplify() for node in self.nodes if not node.is_null)
        if len(nodes) == 1:
            return nodes[0]
        return self

    def children(self) -> Sequence["GrammarNode"]:
        return self.nodes

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.join(state, self, **kwargs)


@dataclass(frozen=True)
class RepeatNode(GrammarNode):
    node: GrammarNode
    min: int
    max: Optional[int]

    @property
    def is_null(self) -> bool:
        return self.node.is_null or self.min == self.max == 0

    def __post_init__(self):
        if self.min < 0:
            raise ValueError("min must be >= 0")
        if self.max is not None and self.max < self.min:
            raise ValueError("max must be >= min")

    def children(self) -> Sequence["GrammarNode"]:
        return (self.node,)

    def simplify(self) -> GrammarNode:
        return RepeatNode(self.node.simplify(), self.min, self.max)

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.repeat(state, self, **kwargs)


@dataclass(frozen=True)
class SubstringNode(GrammarNode):
    chunks: tuple[str, ...]

    @property
    def is_terminal(self) -> bool:
        # this can be used as part of bigger regexes
        return True

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.substring(state, self, **kwargs)


# This creates a name for the given grammar node (value), which can be referenced
# via RuleRefNode (or directly).
# In Lark syntax this results in approx. "{name}: {value}"
# This can either Lark rule (non-terminal) or terminal definition
# (meaning name can be upper- or lowercase).
@dataclass(frozen=True)
class RuleNode(GrammarNode):
    name: str
    value: GrammarNode
    capture: Optional[str] = None
    list_append: bool = False
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None
    stop: Optional[Union[RegexNode, LiteralNode]] = None
    suffix: Optional[LiteralNode] = None
    stop_capture: Optional[str] = None

    def __post_init__(self) -> None:
        if (
            self.temperature is not None
            or self.max_tokens is not None
            or self.stop is not None
            or self.suffix is not None
            or self.stop_capture is not None
        ) and not (self.value.is_terminal or isinstance(self.value, BaseSubgrammarNode)):
            raise ValueError(
                "RuleNode is not terminal, so it cannot have a temperature, max_tokens, or stop condition"
            )

    @property
    def is_terminal(self) -> bool:
        return (
            (
                self.capture is None
                and self.temperature is None
                and self.max_tokens is None
                and self.stop is None
                and self.suffix is None
                and self.stop_capture is None
            )
            and self.value.is_terminal
            and not isinstance(self.value, BaseSubgrammarNode)
        )

    def children(self) -> Sequence["GrammarNode"]:
        return (self.value,)

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.rule(state, self, **kwargs)


@dataclass(frozen=True, eq=False)
class RuleRefNode(GrammarNode):
    target: Optional[RuleNode] = field(default=None, init=False)

    def set_target(self, target: RuleNode) -> None:
        if self.target is not None:
            raise ValueError("RuleRefNode target already set")
        # Side-step frozen=True to set target
        object.__setattr__(self, "target", target)

    @property
    def is_terminal(self) -> bool:
        # RuleRefNode should only ever be used to enable recursive rule definitions,
        # so it should never be terminal.
        return False

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        if self.target is None:
            raise ValueError("RuleRefNode target not set")
        return client.rule(state, self.target)


@dataclass(frozen=True)
class BaseSubgrammarNode(GrammarNode):
    name: str

    @property
    def is_terminal(self) -> bool:
        return False


@dataclass(frozen=True)
class SubgrammarNode(BaseSubgrammarNode):
    body: GrammarNode
    skip_regex: Optional[str] = None

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.subgrammar(state, self, **kwargs)


@dataclass(frozen=True, eq=False)
class JsonNode(BaseSubgrammarNode):
    schema: dict[str, Any]

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.json(state, self, **kwargs)


@dataclass(frozen=True, eq=False)
class LarkNode(BaseSubgrammarNode):
    lark_grammar: str

    def _run(self, client: "Client[S]", state: S, **kwargs) -> Iterator[OutputAttr]:
        return client.lark(state, self, **kwargs)


class LLSerializer:
    def __init__(self):
        self.grammars: dict[str, Union[JsonGrammar, LarkGrammar]] = {}
        self.names: dict[BaseSubgrammarNode, str] = {}

    def serialize(self, node: GrammarNode) -> LLGrammar:
        if isinstance(node, BaseSubgrammarNode):
            self.visit(node)
        else:
            self.visit(SubgrammarNode("main", node))
        return LLGrammar(grammars=[self.grammars[name] for name in self.names.values()])

    def visit(self, node: BaseSubgrammarNode) -> str:
        if node in self.names:
            return self.names[node]

        name = node.name
        names = set(self.names.values())
        if name in names:
            i = 1
            while f"{name}_{i}" in names:
                i += 1
            name = f"{name}_{i}"

        if isinstance(node, SubgrammarNode):
            # Important: insert name BEFORE visiting body to avoid infinite recursion
            self.names[node] = name
            lark_grammar = LarkSerializer(self).serialize(node.body)
            if node.skip_regex:
                lark_grammar += f"\n%ignore /{node.skip_regex}/"
            self.grammars[name] = LarkGrammar(name=name, lark_grammar=lark_grammar)

        elif isinstance(node, JsonNode):
            self.names[node] = name
            self.grammars[name] = JsonGrammar(name=name, json_schema=node.schema)

        elif isinstance(node, LarkNode):
            self.names[node] = name
            self.grammars[name] = LarkGrammar(name=name, lark_grammar=node.lark_grammar)

        else:
            raise TypeError(f"Unknown subgrammar type: {node}")

        return name


class LarkSerializer:

    def __init__(self, ll_serializer: LLSerializer):
        self.ll_serializer = ll_serializer

        self.rules: dict[str, str] = {}
        self.names: dict[RuleNode, str] = {}

    def serialize(self, node: GrammarNode) -> str:
        if isinstance(node, RuleNode) and node.name == "start":
            self.visit(node)
        else:
            self.visit(RuleNode("start", node))

        res = "%llguidance {}\n\n"
        if "start" not in self.rules:
            assert "START" in self.rules
            res += "start: START\n"

        prev_nl = True
        for name in self.names.values():
            s = self.rules[name]
            if not prev_nl and "\n" in s:
                res += "\n"
            res += s + "\n"
            prev_nl = "\n" in s
            if prev_nl:
                res += "\n"

        return res

    def visit(self, node: GrammarNode, top=False) -> str:
        if isinstance(node, BaseSubgrammarNode):
            return f"@{self.ll_serializer.visit(node)}"

        if isinstance(node, RuleNode):
            if node in self.names:
                return self.names[node]

            name = self.normalize_name(node.name, node.is_terminal)
            names = set(self.names.values())
            if name in names:
                i = 1
                while f"{name}_{i}" in names:
                    i += 1
                name = f"{name}_{i}"
            self.names[node] = name
            res = name
            attrs = []
            if node.capture is not None:
                if node.capture != node.name or node.list_append:
                    capture_name = node.capture
                    if node.list_append:
                        capture_name = f"__LIST_APPEND:{capture_name}"
                    attrs.append(f"capture={json.dumps(capture_name)}")
                else:
                    attrs.append("capture")
            if node.temperature is not None:
                attrs.append(f"temperature={node.temperature}")
            if node.max_tokens is not None:
                attrs.append(f"max_tokens={node.max_tokens}")
            if node.stop:
                attrs.append(f"stop={self.visit(node.stop)}")
            if node.suffix:
                attrs.append(f"suffix={self.visit(node.suffix)}")
            if node.stop_capture:
                attrs.append(f"stop_capture={json.dumps(node.stop_capture)}")
            if attrs:
                res += f"[{', '.join(attrs)}]"
            res += ": " + self.visit(node.value.simplify(), top=True)
            self.rules[name] = res
            return name

        if node.is_null:
            return '""'

        if isinstance(node, LiteralNode):
            return json.dumps(node.value)

        if isinstance(node, RegexNode):
            rx = node.regex
            if rx is None:
                rx = "(?s:.*)"
            return self.regex(rx)

        if isinstance(node, SelectNode):
            if top:
                return "\n     | ".join(self.visit(alt) for alt in node.alternatives)
            else:
                return "(" + " | ".join(self.visit(alt) for alt in node.alternatives) + ")"

        if isinstance(node, JoinNode):
            return " ".join(self.visit(n) for n in node.nodes if not n.is_null)

        if isinstance(node, RepeatNode):
            inner = self.visit(node.node)
            if isinstance(node.node, (JoinNode, RepeatNode)):
                inner = f"({inner})"
            if (node.min, node.max) == (0, None):
                return f"{inner}*"
            if (node.min, node.max) == (1, None):
                return f"{inner}+"
            if (node.min, node.max) == (0, 1):
                return f"{inner}?"
            if node.max is None:
                return f"{inner}{{{node.min},}}"
            return f"{inner}{{{node.min},{node.max}}}"

        if isinstance(node, SubstringNode):
            return f'%regex {json.dumps({"substring_chunks": node.chunks}, indent=2)}'

        if isinstance(node, RuleRefNode):
            if node.target is None:
                raise ValueError("RuleRefNode has no target")
            return self.visit(node.target)

        raise TypeError(f"Unknown node type: {node}")

    def normalize_name(self, name: str, terminal: bool) -> str:
        new_name = name.replace("-", "_")
        # convert fooBar_Baz to foo_Bar_Baz
        new_name = re.sub(r"([a-z])([A-Z])", r"\1_\2", new_name)
        if terminal:
            new_name = new_name.upper()
        else:
            new_name = new_name.lower()
        return new_name

    def regex(self, pattern: str) -> str:
        escaped = re.sub(r"(?<!\\)/", r"\/", pattern).replace("\n", "\\n")
        return f"/{escaped}/"



from .._guidance import guidance
from .._grammar import capture as grammar_capture, GrammarNode
from ._block import block

@guidance(stateless=lambda *args, **kwargs: isinstance(args[0], GrammarNode))
def capture(lm, value, name):
    if isinstance(value, GrammarNode):
        return lm + grammar_capture(value, name)
    else:
        with block(name):
            lm += value
        return lm



""" Poor man's exchanges for routing messages. """

from collections import defaultdict
from typing import Callable
from ..visual import GuidanceMessage
import re
import logging

logger = logging.getLogger(__name__)

DEFAULT_TOPIC = "/default"
WILDCARD_PATTERN = r".*"


class TopicExchange:
    """ Queue-less topic exchange for routing messages.

    This is not as comprehensive as a full distributed topic exchange.
    It is specific to a single process, with no queues and less generalized routing keys.
    """

    def __init__(self):
        """ Initializes."""
        self._observers = defaultdict(list)

    def subscribe(self, callback: Callable[[GuidanceMessage], None], topic_pat: str = WILDCARD_PATTERN) -> None:
        """ Subscribes to incoming messages.

        Args:
            callback: Callback to handle incoming messages.
            topic_pat: Topic to notify.
        """
        logger.debug(f"EXCHANGE:pre_subscribe:{self._observers[topic_pat]}")
        self._observers[topic_pat].append(callback)
        logger.debug(f"EXCHANGE:post_subscribe:{self._observers[topic_pat]}")

    def unsubscribe(self, callback: Callable[[GuidanceMessage], None], topic_pat: str = WILDCARD_PATTERN) -> None:
        """ Unsubscribes from incoming messages.

        Args:
            callback: Callback to remove.
            topic_pat: Topic pattern.
        """
        logger.debug(f"EXCHANGE:pre_unsubscribe:{self._observers[topic_pat]}")
        try:
            self._observers[topic_pat].remove(callback)
        except ValueError as _:
            logger.warning(f"EXCHANGE:cb at '{topic_pat}' already removed.")
        logger.debug(f"EXCHANGE:post_unsubscribe:{self._observers[topic_pat]}")

        if len(self._observers[topic_pat]) == 0:
            logger.debug(f"EXCHANGE:delete_entry:{topic_pat}")
            del self._observers[topic_pat]

    def publish(self, message: GuidanceMessage, topic: str = DEFAULT_TOPIC):
        """ Notifies all subscribers to topic pattern of an incoming message.

        Args:
            message: Incoming message.
            topic: Topics to notify.
        """
        # logger.debug(f"EXCHANGE:publish:{message}")
        for obs_topic_pat, observers in self._observers.items():
            if re.match(obs_topic_pat, topic):
                for observer in observers:
                    observer(message)


__all__ = ["TopicExchange"]
